import os

from langchain_openai import AzureChatOpenAI, ChatOpenAI

from parsera.utils import singleton

from langchain.chat_models.base import BaseChatModel
from typing import List, Optional, Any
from langchain_core.outputs.chat_generation import ChatGeneration
from langchain_core.callbacks.manager import CallbackManagerForLLMRun, AsyncCallbackManagerForLLMRun
from langchain_core.outputs.chat_result import ChatResult
from langchain.chat_models.base import BaseMessage
from langchain_core.messages.ai import AIMessage
from functools import partial
import asyncio

@singleton
class AzureModel(AzureChatOpenAI):
    def __init__(self, *args, **kwargs):
        super().__init__(
            azure_endpoint=os.getenv("AZURE_GPT_BASE_URL"),
            openai_api_version="2023-05-15",
            deployment_name=os.getenv("AZURE_GPT_DEPLOYMENT_NAME"),
            openai_api_key=os.getenv("AZURE_GPT_API_KEY"),
            openai_api_type="azure",
            temperature=0.0,
            timeout=120,
            *args,
            **kwargs,
        )


@singleton
class GPT4oMiniModel(ChatOpenAI):
    def __init__(self, *args, **kwargs):
        super().__init__(
            model="gpt-4o-mini",
            temperature=0.0,
            timeout=120,
            *args,
            **kwargs,
        )

@singleton
class HuggingFaceModel(BaseChatModel):
    """
    Model for using HuggingFace pipeline.

    Imports embeded in the class to avoid all users 
    from having to install necessary dependencies 
    for this model.
    """

    pipeline: Any
    """HuggingFace Transformers Pipeline"""

    def _generate(
    self,
    messages: List[BaseMessage],
    stop: Optional[List[str]] = None,
    run_manager: Optional[CallbackManagerForLLMRun] = None,
    **kwargs: Any,
    ) -> ChatResult:
        # Implement the logic for generating the response using the HuggingFace model
        output_str = self._call(messages, stop=stop, run_manager=run_manager, **kwargs)
        message = AIMessage(content=output_str)
        generation = ChatGeneration(message=message)
        return ChatResult(generations=[generation])
    
    def _call(
    self,
    messages: List[BaseMessage],
    stop: Optional[List[str]] = None,
    run_manager: Optional[CallbackManagerForLLMRun] = None,
    **kwargs: Any,
    ) -> str:
        from transformers import Pipeline

        if not isinstance(self.pipeline, Pipeline):
            raise ValueError("pipeline must be a HuggingFace Pipeline")
        
        pipe = self.pipeline

        messages = [
            {"role": "system", "content": messages[0].content},
            {"role": "user", "content": messages[1].content}
        ]

        response = pipe(messages)
        generated_text = response[0]['generated_text'][-1]['content']
        
        return generated_text
    
    async def _agenerate(
    self,
    messages: List[BaseMessage],
    stop: Optional[List[str]] = None,
    run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
    **kwargs: Any,
    ) -> ChatResult:
        func = partial(
            self._generate, messages, stop=stop, run_manager=run_manager, **kwargs
        )
        return await asyncio.get_event_loop().run_in_executor(None, func)
    
    @property
    def _llm_type(self) -> str:
        """Get the type of language model used by this chat model."""
        return "custom"



import asyncio
import enum

from langchain_core.language_models import BaseChatModel

from parsera.engine.model import GPT4oMiniModel
from parsera.engine.simple_extractor import (
    ItemExtractor,
    ListExtractor,
    TabularExtractor,
)
from parsera.page import PageLoader, fetch_page_content


class Parsera:
    class ExtractorType(enum.Enum):
        LIST = ListExtractor
        TABULAR = TabularExtractor
        ITEM = ItemExtractor

    def __init__(
        self,
        model: BaseChatModel | None = None,
        extractor: ExtractorType = ExtractorType.TABULAR,
    ):
        if model is None:
            self.model = GPT4oMiniModel()
        else:
            self.model = model
        self.extractor = extractor
        self.loader = PageLoader()

    async def _run(
        self, url: str, elements: dict, proxy_settings: dict | None = None
    ) -> dict:
        content = await self.loader.load_content(url=url, proxy_settings=proxy_settings)
        extractor_instance = self.extractor.value(
            elements=elements, model=self.model, content=content
        )
        result = await extractor_instance.run()
        return result

    def run(self, url: str, elements: dict, proxy_settings: dict | None = None) -> dict:
        return asyncio.run(
            self._run(url=url, elements=elements, proxy_settings=proxy_settings)
        )

    async def arun(
        self, url: str, elements: dict, proxy_settings: dict | None = None
    ) -> dict:
        return await self._run(
            url=url, elements=elements, proxy_settings=proxy_settings
        )



import warnings
from typing import Literal, TypedDict

from playwright.async_api import (
    Browser,
    BrowserContext,
    Page,
    Playwright,
    async_playwright,
)
from playwright_stealth import stealth_async


class ProxySettings(TypedDict, total=False):
    server: str
    bypass: str | None = None
    username: str | None = None
    password: str | None = None


class PageLoader:
    def __init__(
        self,
        browser: Literal["firefox", "chromium"] = "firefox",
    ):
        self._browser_id = browser
        self.playwright: Playwright | None = None
        self.browser: Browser | None = None
        self.context: BrowserContext | None = None
        self.page: Page | None = None

    async def new_browser(self) -> None:
        if not self.playwright:
            self.playwright = await async_playwright().start()

        if self.browser:
            await self.browser.close()

        if self._browser_id == "firefox":
            self.browser = await self.playwright.firefox.launch(headless=True)
        else:
            self.browser = await self.playwright.chromium.launch(headless=True)

    async def load_content(
        self,
        url: str,
        proxy_settings: ProxySettings | None = None,
        new_browser: bool = True,
        load_state: Literal[
            "domcontentloaded", "load", "networkidle"
        ] = "domcontentloaded",
    ) -> None:
        if new_browser:
            await self.new_browser()
        self.context = await self.browser.new_context(proxy=proxy_settings)

        self.page = await self.context.new_page()
        await stealth_async(self.page)
        # Navigate to the URL
        # await page.route("**/*.{png,jpg,jpeg}", lambda route: route.abort()) # Can speed up requests
        await self.page.goto(url)
        await self.page.wait_for_load_state(load_state)
        return await self.page.content()

    async def close(self) -> None:
        if self.playwright:
            await self.browser.close()
            self.playwright.stop()


async def fetch_page_content(
    url: str,
    proxy_settings: ProxySettings | None = None,
    browser: str = "firefox",
) -> str:
    warnings.warn(
        "fetch_page_content is deprecated and will be removed",
        DeprecationWarning,
    )
    async with async_playwright() as p:
        # Launch the browser
        if browser == "firefox":
            browser = await p.firefox.launch(headless=True)
        else:
            browser = await p.chromium.launch(headless=True)
        # Open a new browser context
        context = await browser.new_context(proxy=proxy_settings)
        # Open a new page
        page = await context.new_page()
        await stealth_async(page)

        # Navigate to the URL
        # await page.route("**/*.{png,jpg,jpeg}", lambda route: route.abort()) # Can speed up requests
        await page.goto(url)

        # Wait for the content to be dynamically loaded
        await page.wait_for_load_state("domcontentloaded")
        # Get the page content
        content = await page.content()

        # Close the browser
        await browser.close()

        return content
